from common import *
import Gnuplot

class GnuplotBackend(BaseClass):
    """
    Backend using the Gnuplot.py Python module.
    
    Tip:
        To close figure window, press <q> when active
    """

    def __init__(self):
        BaseClass.__init__(self)
        self.init()
        
    def init(self):
        #self._g = self.figs[self._attrs['curfig']]._g
        #self.replot = self._g.replot  # Replot all plotobjects (python ojects)
        #self.clear = self._g.clear    # Clear all plotobjects (python objects)
        #self.reset = self._g.reset    # Reset Gnuplot, i.e. blank screen
        # Necessary to add a Gnuplot Session  as _g to the Figure instance
        # self._g will now point to the correct intance saved as _g in curfig
        self.figure(self._attrs['curfig'])

        # Convert table for formatstrings
        self._colors = {'': None,
                        'k':-1,  # Black
                        'r': 1,  # Red 
                        'g': 2,  # Green
                        'b': 3,  # Blue
                        'm': 4,  # Magenta-->Purple
                        'c': 5,  # Cyan-->Aqua
                        'w':-1,  # White-->Black, white on a white is not good
                        'y': 7,  # Yellow-->Orange    
                        }
        
        self._line_types = {'': None,           # No line --> point
                            '-': 'lines',       # Solid 
                            ':': 'lines',       # Dotted line -->solid
                            '-.':'lines',       # Dot-dashed line -->solid
                            '--':'lines',       # Dashed line -->solid
                            }
        
        self._point_types = {'':  None,
                             #'v': 1, # Down-triangle -->Diamond
                             #'+': 2, # Plus
                             #'s': 3, # Square
                             #'x': 4, # Cross
                             #'^': 5, # Triangle
                             #'.': 6, # Dot
                             #'o': 1, # Circle-->Diamond  
                             '.': 0, # point
                             'o': 6, # circle
                             '+': 1, # plus
                             'x': 2, # x-mark
                             '*': 3, # star
                             's': 4, # square
                             'd': 12,# diamond
                             'v': 10,# triangle (down)
                             '^': 8, # triangle (up)
                             '<': 10,# triangle (left) --> (down)
                             '>': 10,# triangle (right) --> (down)
                             'p': 5, # pentagram --> square
                             'h': 5  # hexagram --> square
                             }
  
    def _createGnuplotData(self):
        fig = self._figs[self._attrs['curfig']]
        fig._gdata = [] 
        for item in fig.lines:
            linetype = self._line_types[item.attrs['line_type']]
            color = self._colors[item.attrs['color']]
            pointtype = self._point_types[item.attrs['marker']]
            try: pointsize = int(item.attrs['pointsize']) # "" gives error
            except: pointsize = 1
            # ex: set linestyle 2 lt 1 lw 3 pt 6
            # Can't use linetype/lt since it's not supported \
            # for all versions of gnuplot (lt indicated dotted dashed etc.) 

            withstring = None
            if color: 
                if linetype == None:
                    if pointtype:
                        withstring = "points lt %d pt %d ps %d "\
                                     %(color, pointtype, pointsize)
                    else:
                        withstring = "lines lt %d" %color # Lines is default
                elif linetype == 'lines':
                    if pointtype == None: # Pointtype is not set
                        withstring = "lines lt %d " %(color)
                    else: 
                        withstring = "linespoints lt %d pt %d"\
                                     %(color, pointtype)
            else: # No color
                if linetype == None:
                    if pointtype:
                        withstring = "points  pt %d ps %d " \
                                     %(pointtype, pointsize)
                    else:
                        withstring = "lines"# No color, no linestyle, no marker
                elif linetype == 'lines':
                    if pointtype == None: #Pointtype is not set
                        withstring = "lines" 
                    else: 
                        withstring = "linespoints pt %d" %(pointtype)

            # Create gnuplot data
            fig._gdata.append(Gnuplot.Data(
                item.attrs['x'],
                item.attrs['y'],
                title=item.attrs['legend'],
                with=withstring))
            
        self._gdata=fig._gdata
        #print "withstring", withstring
        #print fig._gdata[0].get_command_option_string()
                
    def figure(self, *args, **kwargs):
        """
        Extension of BaseClass.figure:
        adds a gnuplot instance as figure._g
        then creates a link to it as object._g
        """
        
        BaseClass.figure(self, *args, **kwargs) 
        fig = self._figs[self._attrs['curfig']]

        try:
            fig._g
        except:
            try:
                fig._g = Gnuplot.Gnuplot(persist=1)
                # Plotwindow will now persist
                # To close the gnuplot session run fig._g('quit')
                # Python will only remove the binding to the session and not
                # stop it when _g is deleted            
                # This is due to the persist=1 parameter
            except:
                fig._g = Gnuplot.Gnuplot() # Persist is not supported under win

            self._g = fig._g # Creates link for faster access
        
    def _replot(self):
        self._createGnuplotData()
        fig = self._figs[self._attrs['curfig']]
        scale = fig.attrs['scale']
                
        if scale == 'loglog':
            self._g('set logscale x')
            self._g('set logscale y')
            self._g('set autoscale')
        elif scale == 'logx':
            self._g('set logscale x')
            self._g('set nologscale y')
            self._g('set autoscale')
        elif scale == 'logy':
            self._g('set logscale y')
            self._g('set nologscale x')
            self._g('set autoscale')
        elif scale == 'linear':
            self._g('set nologscale y')
            self._g('set nologscale x')
            self._g('set autoscale')

        # Set axis range
        fail = False
        for item in ['xmin', 'xmax']:
            if  isinstance(fig.attrs[item], (float, int)):#and fig.attrs[item]:
                pass
            else:
                fail = True
        
        if not fail:
            self._g('set xrange[%g:%g]' % (fig.attrs['xmin'],
                                           fig.attrs['xmax']))
        fail = False
        for item in ['ymin', 'ymax']:
            if  isinstance(fig.attrs[item], (float, int)):#and fig.attrs[item]:
                pass
            else:
                fail = True
        if not fail:       
            self._g('set yrange[%g:%g]' % (fig.attrs['ymin'],
                                           fig.attrs['ymax']))

        fail = False
        for item in ['zmin', 'zmax']:
            if  isinstance(fig.attrs[item], (float, int)):#and fig.attrs[item]:
                pass
            else:
                fail = True

        if not fail:       
            self._g('set zrange[%g:%g]' % (fig.attrs['zmin'],
                                           fig.attrs['zmax']))

        # Need to curry _gdata since they now are a list or
        self._g.plot(fig._gdata[0])
        if len(fig._gdata)>1:
            for item in fig._gdata[1:]:
                self._g.replot(item)
        
        self._g.xlabel(fig.attrs['xlabel'])
        self._g.ylabel(fig.attrs['ylabel'])
        self._g.title(fig.attrs['title'])
       
        # Why is this now necessary to get titles and labels right?
        self._g.replot()

    def closeall(self):
        """close figurewindows and stop gnuplot """
        for key  in self._figs.keys():
            self._figs[key]._g('quit')
        del self._g
        self._figs = {1:Figure()}
        self._figs[1]._g = Gnuplot.Gnuplot()
        self._g = self._figs[1]._g

    def hardcopy(self, filename='gpout.ps', **kwargs):
        if 'color' in kwargs:
            self.set(color=kwargs['color'])
            
        color = self.get('color')
        
        if self.get('show'): # Ok to display to screen
            self._replot()
            self._g.hardcopy(filename, color=color, mode='eps',
                             enhanced=True,
                             fontname='Times-Roman', fontsize=20)
        else: # Manually set terminal to ps and don't show windows
            if color:
                colortype = 'color'
            else:
                colortype = 'monochrome'
                        
            # Create a new Gnuplot instance only for now
            self._g = Gnuplot.Gnuplot()
            self._g('set term postscript eps %s' %colortype)
            self._g('set output  "%s"' %filename)
            self._replot()
            self._g.hardcopy(filename, color=color, mode='eps',
                             enhanced=True,
                             fontname='Times-Roman', fontsize=20)
            self._g('quit')
            #Set _g to the correct instance again
            self._g=self._figs[self.get('curfig')]._g

                      
plt = GnuplotBackend() # Create backend instance
use(plt, globals()) # Export public namespace of plt to globals()

def get_backend():
    return plt._g
